{-

AIF Omega version 2016

Developed under BSD license as part of OFMC

(C) Copyright Alessandro Bruni 2015,2016
(C) Copyright DTU 2010,2016
(C) Copyright Paolo Modesti 2010
(C) Copyright IBM Corp. 2010

All Rights Reserved.

-}

{-
TODO:
- implement uniqueness labels on set declarations
- finish the normalization phase
- implement uniqueness checks
- implement general sanity checks
-}

module AIFOm where

import Types
import Data.Char
import Data.List
import qualified Data.Map as Map

getEnvTuple env k = case find (\(k',_) -> k == k') env of
  Just kv -> kv
  Nothing -> error $ "Identifier " ++ k ++ " not found in " ++ show env

getEnv env k = v where (_, v) = getEnvTuple env k

getSetDec p s =
  case find (\t -> case t of PComp s' _ -> s == s'; _ -> False) (setdecs p) of
    Just t  -> t
    Nothing -> error $ "Set " ++ s ++ " not declared"

isUserType ty =
  case ty of EnumType _ -> True; CountType -> True; Union _ -> True; _ -> False

isFiniteType typeEnv ty =
  case getEnv typeEnv ty of
    EnumType _ -> True
    CountType -> False
    Union ids -> all (isFiniteType typeEnv) ids
    _ -> error "AIFOm.isFiniteType: not a user defined type"

getRange typeEnv ty =
  case getEnv typeEnv ty of
    EnumType range -> range
    CountType -> error "AIFOm.getRange: not applicable to infinite types"
    Union ids -> concatMap (getRange typeEnv) ids
    _ -> error "AIFOm.getRange: not a user defined type"

allComb :: [[a]] -> [[a]]
allComb [] = []
allComb (x:[]) = [[y] | y <- x]
allComb (x:xs) = [y:ys | y <- x, ys <- allComb xs]

-- Remove global variables
-- Remove finite enumerations
normalize :: PASLan -> PASLan
normalize p = p { typedec = userTypesNew, setdecs = setDecsNew }
  where
    -- Phase 1: filter user defined types from global variable
    -- declarations, introduce anonymous types for the latter
    isEnumType (ids, ty) = case ty of EnumType _ -> True; _ -> False
    (userTypes, globalVars) = partition (isUserType . snd) (typedec p)
    anonTypeName id = id ++ "__AnonType"
    anonTypes = concatMap (\ (ids, ty) ->
      case ty of
        Enum names -> [(anonTypeName ids, EnumType names)]
        _ -> []) globalVars
    globalEnv :: [(PIdent, PType)] = concatMap (\ (id, ty) ->
      case ty of
        Enum names -> [(id, UserTypeRef (anonTypeName id))]
        UserTypeRef t -> [(id, UserTypeRef t)]
        Value -> [(id, Value)]
        Untyped -> [(id, Untyped)]
        _ -> error "AIFOm.normalize: expecting variable declaration" ) globalVars
    userTypesNew = userTypes ++ anonTypes
    finiteUserTypes = [(ty, getRange userTypesNew ty) | (ty, _) <- userTypesNew, isFiniteType userTypesNew ty]
    -- Phase 2: desugar set declarations
    setDecsNew = concatMap (\ setdec ->
      case setdec of
        PComp name [] -> [setdec]
        PComp name args -> [PComp (name ++ "__" ++ intercalate "__" range) infiniteArgs | range <- desugaredArgs]
          where
            (finiteArgs, infiniteArgs) = partition ((isFiniteType userTypesNew) . atomName) args
            desugaredArgs = allComb (map ((getRange userTypesNew) . atomName) finiteArgs)) (setdecs p)
    -- normalizeRule rule = [rule]
    --   where
        -- All free variables must be declared globally, are now
        -- attached to the local environment
        -- freeVars = map (getEnvTuple globalEnv) (fvr rule)
        -- (finiteArgs, infiniteArgs) = partition ((isFiniteType userTypesNew) . snd) (ruleEnv rule ++ freeVars)
        -- desugaredArgs = allComb (map ((getRange userTypesNew) . snd) finiteArgs)
        -- finiteVars = map fst finiteArgs
        -- Remove universal quantifiers with finite range on negative
        -- set conditions
        -- trCond (PCond allQ var (PComp set args)) = []
          -- [PCond infiniteAllQ var (subt rho set) | img <- desugaredArgs,
          --  let rho = Map.fromList (zip finiteAllQ img)]
          -- where
          --   PComp _ setArgTypes = getSetDec p set
          --   (finiteSetArgs, infiniteSetArgs) = partition (isFiniteType userTypesNew) setArgTypes
          --   (finiteAllQ, infiniteAllQ) = partition (isFiniteType userTypesNew) allQ 
          --   desugaredSetArgs = allComb (map ((getRange userTypesNew) . atomName) finiteSetArgs)

zero = PComp "0" []

compile :: PASLan -> PASLan
compile p = p { typedec = [], rules = newRules, privfundec = newPrivfundec }
  where
    newRules = nub $ userTypeRules ++ concatMap compileRule (rules p)
    newPrivfundec = [(name, length args) | (PComp name args) <- (setdecs p)] ++
      (privfundec p) ++ [(typeFuncName id, 1) | (id, CountType) <- (typedec p)]
    typeFactName (i:id) = ("is" ++ toUpper i : id)
    typeFuncName (i:id) = (toLower i : id)
    userTypeFactDecs = map (\ (id, ty) -> (typeFactName id, 1)) (typedec p)
    userTypeRules = concatMap
      (\ (id, ty) -> case ty of
          EnumType names ->
            [PRule { ruleName = "userType", ruleEnv = [],
                     lf = [], lp = [], ln = [], fresh = [], rp = [],
                     rf = [PFact (typeFactName id) [PAtom name] | name <- names]}]
          CountType ->
            [PRule { ruleName = "userType", ruleEnv = [],
                     lf = [], lp = [], ln = [], fresh = [], rp = [],
                     rf = [PFact (typeFactName id) [PComp (typeFuncName id) [PAtom "X"]]]}]
          Union types ->
            [PRule { ruleName = "userType", ruleEnv = [],
                     lp = [], ln = [], fresh = [], rp = [],
                     lf = [PFact (typeFactName t) [PAtom "X"]],
                     rf = [PFact (typeFactName id) [PAtom "X"]]} | t <- types]
          _ -> error "AIFOm.compile: no global variables are allowed"
      ) (typedec p)
    compileRule r =
      r { ruleEnv = [],
          lf = typesFacts ++ (subfs lambda (lf r)),
          rf = (subfs rho (rf r)) ++ timpliesFacts,
          lp = [], ln = [], fresh = [], rp = [] } : timpliesRules
      where
        left s x =
          case find (\(PCond _ x' (PComp s' _)) -> s == s' && x == x') (lp r) of
            Just (PCond _ _ (PComp _ args)) -> PComp s args
            Nothing ->
              case find (\(PCond _ x' (PComp s' _)) -> s == s' && x == x') (ln r) of
                Just (PCond _ _ (PComp _ args)) -> zero
                Nothing -> PAtom ("E__" ++ s ++ "__" ++ x)
        right s x =
          case find (\(PCond _ x' (PComp s' _)) -> s == s' && x == x') (rp r) of
            Just (PCond _ _ (PComp _ args)) -> PComp s args
            Nothing ->
              case (elem x (fresh r), left s x) of
                (False, PAtom e_i_x) -> PAtom e_i_x
                _ -> zero
        leftVal x = PComp "val" [left s x | PComp s _ <- setdecs p]
        rightVal x = PComp "val" [right s x | PComp s _ <- setdecs p]
        lambda = Map.fromList [(x, leftVal x) | (x, Value) <- ruleEnv r]
        rho = Map.fromList
          ([(x, rightVal x) | (x, Value) <- ruleEnv r] ++
           [(x, rightVal x) | x <- fresh r])
        timpliesFacts =
          [PFact "timplies" [leftVal x, rightVal x] |
           (x, ty) <- ruleEnv r, not (elem x (fresh r)), ty == Value, leftVal x /= rightVal x]
        typesFacts =
          [PFact (typeFactName ty) [PAtom x] | (x, UserTypeRef ty) <- ruleEnv r]
        timpliesRules =
          [PRule { ruleName = "timplies", ruleEnv = [],
                   lf = typesFacts ++
                        [subf sigma_1 fact,
                         PFact "timplies" [PAtom "Val_1", PAtom "Val_2"]],
                   rf = [subf sigma_2 fact],
                   lp = [], ln = [], fresh = [], rp = [] }
          | fact <- rf r, x <- fvf fact,
            getEnv (ruleEnv r) x == Value,
            let sigma_1 = Map.fromList [(x, PAtom "Val_1")],
            let sigma_2 = Map.fromList [(x, PAtom "Val_2")]]

